/*
 * Copyright (c) 2016-2021 Satlab A/S <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <structmember.h>
#include <fcntl.h>
#include <unistd.h>

#include <satlab/polaris_lora.h>
#include <prop-client/prop_proto.h>
#include <btp/types.h>

#define PROP_CLIENT_SUBSYS_SPEC_CAPSULE	"prop_client.subsys_spec"

/* Polaris LoRa object */
typedef struct {
	PyObject_HEAD
	uint8_t node;
	uint32_t timeout;
	PyObject *prop;
	PyObject *btp;
	PyObject *csp;
} PolarisLoRaObject;

/* Polaris LoRa Frame object */
typedef struct {
	PyObject_HEAD
	uint32_t seq;
	int32_t frequency_offset;
	int16_t symbol_power;
	int16_t noise_power;
	uint8_t spreading_factor;
	uint8_t code_rate;
	uint8_t payload_crc_enabled;
	uint64_t time;
	PyObject *data;
} PolarisLoRaFrameObject;

static PyTypeObject PolarisLoRaFrameType;

/* Callback mapping */
typedef struct {
	PyObject *cb;
	PyObject *cbarg;
} PolarisLoRa_store_mirror_recv_cbmap;

static int PolarisLoRa_store_mirror_recv_cb(sl_polaris_lora_store_frame_t *frame, void *arg)
{
	int ret = 0;

	PyGILState_STATE gstate;
	PolarisLoRa_store_mirror_recv_cbmap *cbmap = arg;

	/* Grab GIL before calling back into Python */
	gstate = PyGILState_Ensure();

	if (frame != NULL) {
		PyObject *cbret, *arglist, *frameobj, *objargs, *databytes;

		/* Create data bytes object */
		databytes = PyBytes_FromStringAndSize((const char *)frame->data, frame->length);
		/* Create constructor args */
		objargs = Py_BuildValue("IihhbbbkO",
					frame->seq,
					frame->frequency_offset,
					frame->symbol_power,
					frame->noise_power,
					frame->spreading_factor,
					frame->code_rate,
					frame->payload_crc_enabled,
					frame->time,
					databytes);

		/* Call the class object */
		frameobj = PyObject_CallObject((PyObject *) &PolarisLoRaFrameType, objargs);

		/* Build argument list */
		arglist = Py_BuildValue("(OO)", frameobj, cbmap->cbarg);

		/* Call Python callback */
		cbret = PyObject_CallObject(cbmap->cb, arglist);
		Py_DECREF(databytes);
		Py_DECREF(objargs);
		Py_DECREF(frameobj);
		Py_DECREF(arglist);

		ret = PyLong_AsLong(cbret);
		Py_DECREF(cbret);
	}

	/* Check for ctrl+c */
	PyErr_CheckSignals();
	if (PyErr_Occurred()) {
		if (PyErr_ExceptionMatches(PyExc_KeyboardInterrupt)) {
			ret = -EINTR;
		} else {
			ret = -EIO;
		}
	}

	/* Release GIL */
	PyGILState_Release(gstate);

	return ret;
}

static PyObject *PolarisLoRa_store_mirror_recv(PolarisLoRaObject *self, PyObject *args, PyObject *kwds)
{
	int ret;

	static const char *kwlist[] = {"cb", "cbarg", "timeout", NULL};
	PolarisLoRa_store_mirror_recv_cbmap cbmap = {
		.cbarg = Py_None,
	};
	uint32_t timeout = self->timeout;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|OI", (char **)kwlist, &cbmap.cb, &cbmap.cbarg, &timeout))
		return NULL;

	if (!PyCallable_Check(cbmap.cb)) {
		PyErr_SetString(PyExc_TypeError, "cb must be callable");
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	ret = sl_polaris_lora_store_mirror_recv(self->node, timeout, PolarisLoRa_store_mirror_recv_cb, &cbmap);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		if (ret != -EINTR)
			PyErr_Format(PyExc_TimeoutError, "Failed to request store mirror: %d\n", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static PyObject *PolarisLoRa_sample_file(PolarisLoRaObject *self, PyObject *args, PyObject *kwds)
{
	int ret;
	int fd, nbytes;
	uint32_t *sample_mem;
	uint16_t repsiz = 256;
	uint32_t timeout = 5000;
	size_t samples = 8192;

	static const char *kwlist[] = {"filename", "timeout", "samples", "repsiz", NULL};
	const char *file;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|IIh", (char **)kwlist, &file, &timeout, &samples, &repsiz))
		return NULL;

	fd = open(file, O_WRONLY | O_CREAT | O_TRUNC,
		  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd < 0) {
		PyErr_Format(PyExc_IOError, "Failed to open output file\n");
		return NULL;
	}

	sample_mem = calloc(samples, sizeof(*sample_mem));
	if (!sample_mem) {
		PyErr_Format(PyExc_MemoryError, "Failed to allocate memory\n");
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	ret = sl_polaris_lora_sample_trigger_buffer(self->node, timeout, 0, repsiz, samples, 0, 0,sample_mem, samples * sizeof(*sample_mem));
	Py_END_ALLOW_THREADS

	if (ret < 0) {
		free(sample_mem);
		printf("Failed to trigger sample: %s\n", sl_polaris_lora_strerror(-ret));
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	nbytes = write(fd, sample_mem, samples * sizeof(*sample_mem));
	Py_END_ALLOW_THREADS

	if (nbytes != samples * sizeof(*sample_mem)) {
		free(sample_mem);
		PyErr_Format(PyExc_IOError, "Failed to write sample to output file\n");
		return NULL;
	}

	free(sample_mem);

	Py_RETURN_NONE;
}

static PyObject *PolarisLoRa_shell_run(PolarisLoRaObject *self, PyObject *args, PyObject *kwds)
{
	int ret;

	static const char *kwlist[] = {"cmd", "timeout", NULL};
	const char *cmd;
	uint32_t timeout = self->timeout;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|I", (char **)kwlist, &cmd, &timeout))
		return NULL;

	Py_BEGIN_ALLOW_THREADS
	ret = sl_polaris_lora_shell_run(self->node, timeout, cmd);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_TimeoutError, "Failed to run command: %d\n", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static int PolarisLoRa_init(PolarisLoRaObject *self, PyObject *args, PyObject *kwds)
{
	int ret = -1;
	uint8_t prop_port = SL_PROP_DEFAULT_PORT;
	uint32_t btp_count = 15;

	self->node = SL_POLARIS_LORA_DEFAULT_ADDRESS;
	self->timeout = 1000;

	static const char *kwlist[] = {"node", "timeout", NULL};
	if (!PyArg_ParseTupleAndKeywords(args, kwds, "|bI", (char **)kwlist, &self->node, &self->timeout))
		goto err;

	/* Get reference to prop client module */
	PyObject *pyprop = PyImport_ImportModule("satlab.prop_client");
	if (!pyprop)
		goto err;

	/* Get reference to PropNode class */
	PyObject *PropNode = PyObject_GetAttrString(pyprop, "PropNode");
	if (!PropNode)
		goto err_prop_node;

	/* Create capsule object with props */
	PyObject *props_capsule = PyCapsule_New((void *)&polaris_lora_props, PROP_CLIENT_SUBSYS_SPEC_CAPSULE, NULL);
	if (!props_capsule)
		goto err_prop_capsule;

	/* Create PropNode object by calling constructor */
	self->prop = PyObject_CallFunction(PropNode, "bbOI", self->node, prop_port, props_capsule, self->timeout);
	if (!self->prop)
		goto err_prop_ctor;

	/* Get reference to CSP module */
	PyObject *pycsp = PyImport_ImportModule("csp");
	if (!pycsp)
		goto err_csp;

	/* Get reference to CSPNode class */
	PyObject *CSPNode = PyObject_GetAttrString(pycsp, "CSPNode");
	if (!CSPNode)
		goto err_csp_node;

	/* Create PropNode object by calling constructor */
	self->csp = PyObject_CallFunction(CSPNode, "b", self->node);
	if (!self->csp)
		goto err_csp_ctor;

	/* Get reference to btp module */
	PyObject *pybtp = PyImport_ImportModule("satlab.btp_client");
	if (!pybtp)
		goto err_btp;

	/* Get reference to BTPNode class */
	PyObject *BTPNode = PyObject_GetAttrString(pybtp, "BTPNode");
	if (!BTPNode)
		goto err_btp_node;

	/* Create BTPNode object by calling constructor */
	self->btp = PyObject_CallFunction(BTPNode, "bIbI", self->node, self->timeout, BTP_DEFAULT_PORT, btp_count);
	if (!self->btp)
		goto err_btp_ctor;


	ret = 0;

err_btp_ctor:
	Py_DECREF(BTPNode);
err_btp_node:
	Py_DECREF(pybtp);
err_btp:
err_csp_ctor:
	Py_DECREF(CSPNode);
err_csp_node:
	Py_DECREF(pycsp);
err_csp:
err_prop_ctor:
	Py_DECREF(props_capsule);
err_prop_capsule:
	Py_DECREF(PropNode);
err_prop_node:
	Py_DECREF(pyprop);
err:
	return ret;
}

static PyMemberDef PolarisLoRa_members[] = {
	{"prop", T_OBJECT_EX, offsetof(PolarisLoRaObject, prop), 0, "PropNode"},
	{"btp",  T_OBJECT_EX, offsetof(PolarisLoRaObject, btp), 0, "BTPNode"},
	{"csp",  T_OBJECT_EX, offsetof(PolarisLoRaObject, csp), 0, "CSPNode"},
	{NULL}  /* Sentinel */
};

static PyMethodDef PolarisLoRa_methods[] = {
	{"store_mirror_recv", (PyCFunction)(void(*)(void))PolarisLoRa_store_mirror_recv, METH_VARARGS | METH_KEYWORDS, "Store mirror."},
	{"sample_file", (PyCFunction)(void(*)(void))PolarisLoRa_sample_file, METH_VARARGS | METH_KEYWORDS, "Sample file."},
	{"shell_run", (PyCFunction)(void(*)(void))PolarisLoRa_shell_run, METH_VARARGS | METH_KEYWORDS, "Run shell command."},
	{NULL}  /* Sentinel */
};

static PyTypeObject PolarisLoRaType = {
	PyVarObject_HEAD_INIT(NULL, 0)
	.tp_name = "polaris_lora_client.PolarisLoRa",
	.tp_doc = "PolarisLoRa object",
	.tp_basicsize = sizeof(PolarisLoRaObject),
	.tp_itemsize = 0,
	.tp_flags = Py_TPFLAGS_DEFAULT,
	.tp_new = PyType_GenericNew,
	.tp_init = (initproc) PolarisLoRa_init,
	.tp_members = PolarisLoRa_members,
	.tp_methods = PolarisLoRa_methods,
};

static int PolarisLoRaFrame_init(PolarisLoRaFrameObject *self, PyObject *args, PyObject *kwds)
{
	static const char *kwlist[] = {"seq",
				 "frequency_offset",
				 "symbol_power",
				 "noise_power",
				 "spreading_factor",
				 "code_rate",
				 "payload_crc_enabled",
				 "time",
				 "data",
				 NULL};

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "IihhbbbLO", (char **)kwlist,
					 &self->seq,
					 &self->frequency_offset,
					 &self->symbol_power,
					 &self->noise_power,
					 &self->spreading_factor,
					 &self->code_rate,
					 &self->payload_crc_enabled,
					 &self->time,
					 &self->data))
		return -1;

	Py_INCREF(self->data);

	return 0;
}

static PyMemberDef PolarisLoRaFrame_members[] = {
	{"seq",                 T_UINT,	     offsetof(PolarisLoRaFrameObject, seq),  0, "seq"},
	{"frequency_offset",    T_INT,       offsetof(PolarisLoRaFrameObject, frequency_offset),  0, "frequency_offset"},
	{"symbol_power",        T_SHORT,     offsetof(PolarisLoRaFrameObject, symbol_power),  0, "symbol_power"},
	{"noise_power",         T_SHORT,     offsetof(PolarisLoRaFrameObject, noise_power),  0, "noise_power"},
	{"spreading_factor",    T_UBYTE,     offsetof(PolarisLoRaFrameObject, spreading_factor),  0, "spreading_factor"},
	{"code_rate",           T_UBYTE,     offsetof(PolarisLoRaFrameObject, code_rate),  0, "code_rate"},
	{"payload_crc_enabled", T_UBYTE,     offsetof(PolarisLoRaFrameObject, payload_crc_enabled),  0, "payload_crc_enabled"},
	{"time",                T_ULONG,     offsetof(PolarisLoRaFrameObject, time), 0, "time"},
	{"data",                T_OBJECT_EX, offsetof(PolarisLoRaFrameObject, data), 0, "data"},
	{NULL}  /* Sentinel */
};

static PyTypeObject PolarisLoRaFrameType = {
	PyVarObject_HEAD_INIT(NULL, 0)
	.tp_name = "polaris_lora_client.PolarisLoRaFrame",
	.tp_doc = "PolarisLoRaFrame object",
	.tp_basicsize = sizeof(PolarisLoRaFrameObject),
	.tp_itemsize = 0,
	.tp_flags = Py_TPFLAGS_DEFAULT,
	.tp_new = PyType_GenericNew,
	.tp_init = (initproc) PolarisLoRaFrame_init,
	.tp_members = PolarisLoRaFrame_members,
};

static struct PyModuleDef polaris_lora_client_module = {
	.m_base = PyModuleDef_HEAD_INIT,
	.m_name = "polaris_lora_client",
	.m_size = -1,
};

PyMODINIT_FUNC PyInit_polaris_lora_client_bindings(void)
{
	if (PyType_Ready(&PolarisLoRaType) < 0)
		return NULL;

	if (PyType_Ready(&PolarisLoRaFrameType) < 0)
		return NULL;

	PyObject *m = PyModule_Create(&polaris_lora_client_module);
	if (!m)
		return NULL;

	/* Add PolarisLoRa object */
	Py_INCREF(&PolarisLoRaType);
	if (PyModule_AddObject(m, "PolarisLoRa", (PyObject *) &PolarisLoRaType) < 0) {
		Py_DECREF(&PolarisLoRaType);
		Py_DECREF(m);
		return NULL;
	}

	/* Add PolarisLoRaFrame object */
	Py_INCREF(&PolarisLoRaFrameType);
	if (PyModule_AddObject(m, "PolarisLoRaFrame", (PyObject *) &PolarisLoRaFrameType) < 0) {
		Py_DECREF(&PolarisLoRaFrameType);
		Py_DECREF(&PolarisLoRaType);
		Py_DECREF(m);
		return NULL;
	}

	PyModule_AddIntConstant(m, "DEFAULT_ADDRESS", SL_POLARIS_LORA_DEFAULT_ADDRESS);

	return m;
}
